# coding: utf-8
#
# BMPC Music Player Client
#
# Copyright (C) 2015 Itaï BEN YAACOV
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


from gi.repository import GObject, Gtk
import sys
import os
import xdg.BaseDirectory
import json
import apsw

import ampd

from .utils.logger import logger


class AppShared(GObject.GObject):
    want_to_connect = GObject.property(type=bool, default=False)
    safe_mode = GObject.property(type=bool, default=False)

    def __init__(self):
        super(AppShared, self).__init__()
        self.uidefs = {}
        self.excepthook_orig, sys.excepthook = sys.excepthook, self.excepthook

        self.ampd_client = ampd.ClientGLib()
        self.ampd_client.connect('client-connected', self.client_connected_cb)
        self.ampd_client.connect('client-disconnected', self.server_disconnected_cb)
        self.ampd_server_properties = ampd.ServerPropertiesGLib(self.ampd_client)
        self.ampd_server_properties_bis = ampd.ServerProperties(self.ampd_client)
        self.ampd_server_properties.connect('server-error', self.server_error_cb)
        for option in self.ampd_server_properties.OPTION_NAMES:
            self.ampd_client.connect('notify::option-' + option, self.notify_option_cb)

        self.ampd_host = None
        self.ampd_port = 6600
        self.ampd_worker_group = ampd.WorkerGroup(self.ampd_client)

        self.plugins = {}

        self.config = {}
        for path in xdg.BaseDirectory.load_config_paths('bmpc'):
            try:
                self.config = self.config or json.load(open(os.path.join(path, 'config.json')))
            except Exception as e:
                logger.warning(e)

        self.config.setdefault('server', {})
        self.config['server'].setdefault('profiles',
                                         [
                                             {
                                                 'name': _("Local host"),
                                                 'host': 'localhost',
                                                 'port': 6600,
                                             },
                                         ])
        self.config['server'].setdefault('profile', _("Local host"))

    def close(self):
        logger.debug("Closing shared")
        s = json.dumps(self.config, sort_keys=True, indent=2, ensure_ascii=False)
        open(os.path.join(xdg.BaseDirectory.save_config_path('bmpc'), 'config.json'), 'w').write(s)
        self.ampd_client.close()
        sys.excepthook = self.excepthook_orig
        del self.plugins
        del self.ampd_client
        del self.ampd_server_properties

    def build_ui(self, uiname):
        if uiname in self.uidefs:
            uidef = self.uidefs[uiname]
        else:
            uidef = open(os.path.join(os.path.dirname(__file__), uiname + '.ui')).read()
            self.uidefs[uiname] = uidef
        builder = Gtk.Builder(translation_domain='bmpc')
        builder.add_from_string(uidef)
        return builder

    def client_connected_cb(self, client):
        logger.info(_("Connected to {} [protocol version {}]").format(self.config['server']['profile'], self.ampd_worker_group.get_protocol_version()))
        self.idle_database()

    @ampd.worker
    def server_disconnected_cb(self, client, reason):
        if reason == client.DISCONNECT_RECONNECT:
            return
        elif reason == client.DISCONNECT_PASSWORD:
            logger.error(_("Invalid password for {}").format(self.config['server']['profile']))
            return
        elif reason == client.DISCONNECT_FAILED_CONNECT:
            logger.error(_("Connection to {} failed").format(self.config['server']['profile']))
        else:
            logger.info(_("Disconnected from {}").format(self.config['server']['profile']))
        if self.want_to_connect:
            reply = yield ampd.CONNECT, ampd.TIMEOUT(1000)
            if ampd.CONNECT not in reply:
                self.ampd_client.connect_to_server(self.ampd_host, self.ampd_port)

    def server_error_cb(self, client, error):
        logger.error(_("Server error: {}").format(error))

    @ampd.worker
    def notify_option_cb(self, properties, param):
        option = param.name.split('-')[1]
        if self.safe_mode:
            yield ampd.commands_[option](0)

    @ampd.worker
    def idle_database(self):
        while True:
            idle = yield ampd.DATABASE, ampd.UPDATE
            status = yield ampd.status()
            if 'updating_db' in status:
                logger.info(_("Updating database"))
            elif ampd.DATABASE in idle:
                logger.info(_("Database changed"))
            else:
                logger.info(_("Database unchanged"))

    @ampd.worker
    def mpd_command_cb(self, caller, *data):
        if not self.ampd_server_properties.state:
            yield ampd.CONNECT
        yield ampd.commands_[caller.get_name()]()

    @ampd.worker
    def play_or_pause_cb(self, caller, *data):
        if not self.ampd_server_properties.state:
            yield ampd.IDLE
        yield ampd.pause(1) if self.ampd_server_properties.state == 'play' else ampd.play()

    @ampd.worker
    def fade_to_next_cb(self, caller, *data):
        N = 30
        T = 7

        if not self.ampd_server_properties.state:
            yield ampd.IDLE
        if self.ampd_server_properties.state != 'play':
            return
        volume = self.ampd_server_properties.volume
        for i in range(N):
            self.ampd_server_properties.volume = volume*(N-i-1)/N
            reply = yield ampd.TIMEOUT(T*1000/N), ampd.PLAYER
            if ampd.PLAYER in reply:
                logger.debug("Bailing out of fade-to-next")
                break
        else:
            yield ampd.next()
        self.ampd_server_properties.volume = volume

    @staticmethod
    def format_time(time):
        time = int(time)
        hours = time // 3600
        minutes = (time // 60) % 60
        seconds = time % 60
        return '{:d}:{:02d}:{:02d}'.format(hours, minutes, seconds) if hours else '{:d}:{:02d}'.format(minutes, seconds)

    def ampd_connect(self, action, state):
        self.want_to_connect = True
        if state:
            action.set_state(state)
            self.config['server']['profile'] = state.get_string()
        name = self.config['server']['profile']
        for profile in self.config['server']['profiles']:
            if profile['name'] == name:
                self.ampd_host = profile['host']
                self.ampd_port = profile['port']
                self.ampd_client.connect_to_server(self.ampd_host, self.ampd_port)
                break

    def ampd_disconnect(self, *args):
        self.want_to_connect = False
        self.ampd_client.disconnect_from_server()

    def excepthook(self, *args):
        logger.error(args[1], exc_info=args)

    @staticmethod
    def sqlite_connection(name):
        connection = apsw.Connection(os.path.join(xdg.BaseDirectory.save_data_path('bmpc'), name + '.sqlite'))
        connection.cursor().execute('PRAGMA FOREIGH_KEYS=ON')
        return connection
